@@ -25,78 +25,13 @@
 #include "BI.h"
 
 
-void Merge(int *permutation, int *source, int *inversion_count, int dim, int n)
-{
-	int *left = (int *)malloc(n * sizeof(int));
-	int *right = (int *)malloc(n * sizeof(int));
-	int *left_source = (int *)malloc(n * sizeof(int));
-	int *right_source = (int *)malloc(n * sizeof(int));
-	int left_index = 0, right_index = 0;
-	int i, half_dim = dim / 2, nleft = half_dim, nright = dim - half_dim;
-	for (i = 0; i < half_dim; i++) {
-		left[i] = permutation[i];
-		left_source[i] = source[i];
-		right[i] = permutation[i + half_dim];
-		right_source[i] = source[i + half_dim];
-	}
-
-	if (nleft < nright) {
-		right[i] = permutation[i + half_dim];
-		right_source[i] = source[i + half_dim];
-	}
-
-	for (i = 0; i < dim; i++) {
-		if ((left_index < half_dim) && (right_index < dim - half_dim)) {
-			if (left[left_index] <= right[right_index]) { // I added "=" in order to support ties
-				permutation[i] = left[left_index];
-				source[i] = left_source[left_index];
-				left_index++;
-			}
-			else {
-				permutation[i] = right[right_index];
-				source[i] = right_source[right_index];
-				inversion_count[source[i]] += (half_dim - left_index);
-				right_index++;
-			}
-		}
-		else {
-			if (left_index < half_dim) {
-				permutation[i] = left[left_index];
-				source[i] = left_source[left_index];
-				left_index++;
-			}
-
-			if (right_index < dim - half_dim) {
-				permutation[i] = right[right_index];
-				source[i] = right_source[right_index];
-				right_index++;
-			}
-		}
-	}
-	free(left);
-	free(right);
-	free(left_source);
-	free(right_source);
-}
-
-
-int Inversions(int *permutation, int *source, int *inversion_count, int dim, int n)
-{
-	if (dim == 1)
-		return 0;
-	else {
-		Inversions(permutation, source, inversion_count, dim / 2, n);
-		Inversions(&permutation[dim / 2], &source[dim / 2], inversion_count, dim - dim / 2, n);
-		Merge(permutation, source, inversion_count, dim, n);
-	}
-	return 0;
-}
-
-
-double Ball_Information(int *n, double **Dx, double **Dy, int **xidx, int **yidx, int *i_perm, int *i_perm_inv, int *weight)
+void Ball_Information(double *bcov_stat, int *n, double **Dx, double **Dy, int **xidx, int **yidx, int *i_perm, int *i_perm_inv)
 {
 	int i, j, k, pi, src, lastpos, *yrank, *isource, *icount, *xy_index, *xy_temp, **xyidx;
-	double pxy, px, py, lastval, rct0 = 0, *xx_cpy, *yy_cpy;
+	double pxy, px, py, lastval, *xx_cpy, *yy_cpy;
+	double bcov_weight0 = 0.0, bcov_weight_prob = 0.0, bcov_weight_hhg = 0.0, bcov_fixed_ball = 0.0;
+	double hhg_ball_num = 0.0, minor_ball_prop = 2 / (*n);
+
 	yrank = (int *)malloc(*n * sizeof(int));
 	isource = (int *)malloc(*n * sizeof(int));
 	icount = (int *)malloc(*n * sizeof(int));
@@ -166,10 +101,14 @@ double Ball_Information(int *n, double **Dx, double **Dy, int **xidx, int **yidx
 			px /= (*n);
 			py /= (*n);
 			pxy /= (*n);
-			if (*weight == 0)
-				rct0 += pow(pxy - px*py, 2);
-			else
-				rct0 += pow(pxy - px*py, 2) / (px*py);
+			bcov_fixed_ball = pow(pxy - px*py, 2);
+			bcov_weight0 += bcov_fixed_ball;
+			bcov_weight_prob += bcov_fixed_ball / (px*py);
+			if (px > minor_ball_prop && py > minor_ball_prop && px != 1 && py != 1)
+			{
+				bcov_weight_hhg += bcov_fixed_ball / ((px - minor_ball_prop)*(1.0 - px + minor_ball_prop)*(py - minor_ball_prop)*(1.0 - py + minor_ball_prop));
+				hhg_ball_num += 1.0;
+			}
 		}
 		pxy = 0;
 		px = 0;
@@ -191,25 +130,34 @@ double Ball_Information(int *n, double **Dx, double **Dy, int **xidx, int **yidx
 		px /= (*n);
 		py /= (*n);
 		pxy /= (*n);
-		if (*weight == 0)
-			rct0 += pow(pxy - px*py, 2);
-		else
-			rct0 += pow(pxy - px*py, 2) / (px*py);
+		bcov_fixed_ball = pow(pxy - px*py, 2);
+		bcov_weight0 += bcov_fixed_ball;
+		bcov_weight_prob += bcov_fixed_ball / (px*py);
+		if (px > minor_ball_prop && py > minor_ball_prop && px != 1 && py != 1)
+		{
+			bcov_weight_hhg += bcov_fixed_ball / ((px - minor_ball_prop)*(1.0 - px + minor_ball_prop)*(py - minor_ball_prop)*(1.0 - py + minor_ball_prop));
+			hhg_ball_num += 1.0;
+		}
 	}
-	rct0 = rct0 / (1.0*(*n)*(*n));
+	bcov_stat[0] = bcov_weight0 / (1.0*(*n)*(*n));
+	bcov_stat[1] = bcov_weight_prob / (1.0*(*n)*(*n));
+	bcov_stat[2] = bcov_weight_hhg / (hhg_ball_num);
+
+	// free memory
 	free(isource);
 	free(icount);
 	free(xy_index);
 	free(yrank);
 	free(xy_temp);
 	free_int_matrix(xyidx, *n, *n);
-	return(rct0);
+	return;
 }
 
 
-double Ball_Information_parallel(int *n, double **Dx, double **Dy, int **xidx, int **yidx, int *i_perm, int *i_perm_inv, int *weight, int *nthread)
+void Ball_Information_parallel(double *bcov_stat, int *n, double **Dx, double **Dy, int **xidx, int **yidx, int *i_perm, int *i_perm_inv, int *nthread)
 {
-	double rct0_value = 0.0;
+	double bcov_weight0 = 0.0, bcov_weight_prob = 0.0, bcov_weight_hhg = 0.0;
+	double minor_ball_prop = 2 / (*n);
 	int i, jjjj, **xyidx;
 	double *xx_cpy, *yy_cpy;
 	xyidx = alloc_int_matrix(*n, *n);
@@ -235,7 +183,8 @@ double Ball_Information_parallel(int *n, double **Dx, double **Dy, int **xidx, i
 #pragma omp parallel
 	{
 		int j, k, pi, src, lastpos, *yrank, *isource, *icount, *xy_index, *xy_temp;
-		double pxy, px, py, lastval, rct0 = 0;
+		double pxy, px, py, lastval;
+		double bcov_weight0_thread = 0.0, bcov_weight_prob_thread = 0.0, bcov_weight_hhg_thread = 0.0, bcov_fixed_ball = 0.0;
 		yrank = (int *)malloc(*n * sizeof(int));
 		isource = (int *)malloc(*n * sizeof(int));
 		icount = (int *)malloc(*n * sizeof(int));
@@ -289,10 +238,13 @@ double Ball_Information_parallel(int *n, double **Dx, double **Dy, int **xidx, i
 				px /= (*n);
 				py /= (*n);
 				pxy /= (*n);
-				if (*weight == 0)
-					rct0 += pow(pxy - px*py, 2);
-				else
-					rct0 += pow(pxy - px*py, 2) / (px*py);
+				bcov_fixed_ball = pow(pxy - px*py, 2);
+				bcov_weight0_thread += bcov_fixed_ball;
+				bcov_weight_prob_thread += bcov_fixed_ball / (px*py);
+				if (px > minor_ball_prop && py > minor_ball_prop && px != 1 && py != 1)
+				{
+					bcov_weight_hhg_thread += bcov_fixed_ball / ((px - minor_ball_prop)*(1.0 - px + minor_ball_prop)*(py - minor_ball_prop)*(1.0 - py + minor_ball_prop));
+				}
 			}
 
 			pxy = 0;
@@ -317,15 +269,23 @@ double Ball_Information_parallel(int *n, double **Dx, double **Dy, int **xidx, i
 			px /= (*n);
 			py /= (*n);
 			pxy /= (*n);
-			if (*weight == 0)
-				rct0 += pow(pxy - px*py, 2);
-			else
-				rct0 += pow(pxy - px*py, 2) / (px*py);
+			bcov_fixed_ball = (pxy - px*py) * (pxy - px*py);
+			bcov_weight0_thread += bcov_fixed_ball;
+			bcov_weight_prob_thread += bcov_fixed_ball / (px*py);
+			if (px > minor_ball_prop && py > minor_ball_prop && px != 1 && py != 1)
+			{
+				bcov_weight_hhg_thread += bcov_fixed_ball / ((px - minor_ball_prop)*(1.0 - px + minor_ball_prop)*(py - minor_ball_prop)*(1.0 - py + minor_ball_prop));
+			}
 		}
 #pragma omp critical
 		{
-			rct0_value += rct0;
+			bcov_weight0 += bcov_weight0_thread;
+			bcov_weight_prob += bcov_weight_prob_thread;
+			bcov_weight_hhg += bcov_weight_hhg_thread;
 		}
+		bcov_stat[0] = bcov_weight0 / (1.0*(*n)*(*n));
+		bcov_stat[1] = bcov_weight_prob / (1.0*(*n)*(*n));
+		bcov_stat[2] = bcov_weight_hhg / (1.0*(*n)*(*n));
 
 		free(isource);
 		free(icount);
@@ -334,27 +294,27 @@ double Ball_Information_parallel(int *n, double **Dx, double **Dy, int **xidx, i
 		free(xy_temp);
 	}
 	free_int_matrix(xyidx, *n, *n);
-	rct0_value = rct0_value / (1.0*(*n)*(*n));
-	return(rct0_value);
+	return;
 }
 
 
-double Ball_Information_wrapper(int *n, double **Dx, double **Dy, int **xidx, int **yidx, int *i_perm, int *i_perm_inv, int *weight, int *nthread)
+void Ball_Information_wrapper(double *bcov_stat, int *n, double **Dx, double **Dy, int **xidx, int **yidx, int *i_perm, int *i_perm_inv, int *nthread)
 {
 	if (*nthread == 1) {
-		return Ball_Information(n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, weight);
+		Ball_Information(bcov_stat, n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv);
 	}
 	else {
-		return Ball_Information_parallel(n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, weight, nthread);
+		Ball_Information_parallel(bcov_stat, n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, nthread);
 	}
+	return;
 }
 
 
-void BI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *R, int *weight, int *thread)
+void BI(double *bcov, double *pvalue, double *x, double *y, int *n, int *R, int *thread)
 {
 	/*  computes RCT(x,y)  */
 	int    i, j, **xidx, **yidx, *i_perm, *i_perm_inv;
-	double **Dx, **Dy, *x_cpy, *y_cpy, RCTV0;
+	double **Dx, **Dy, *x_cpy, *y_cpy;
 
 	Dx = alloc_matrix(*n, *n);
 	Dy = alloc_matrix(*n, *n);
@@ -365,6 +325,7 @@ void BI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *
 	x_cpy = (double *)malloc(*n * sizeof(double));
 	y_cpy = (double *)malloc(*n * sizeof(double));
 
+	// convert distance vector to distance matrix:
 	vector2matrix(x, Dx, *n, *n, 1);
 	vector2matrix(y, Dy, *n, *n, 1);
 
@@ -379,6 +340,9 @@ void BI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *
 		i_perm_inv[i] = i;
 	}
 
+	// sort each row of Dx and Dy
+	// computational complexity: O(n^2 * logn)
+	// xidx, yidx is index of each row after sorted
 	for (i = 0; i<(*n); i++)
 	{
 		// copy site to x_cpy and y_cpy
@@ -390,17 +354,30 @@ void BI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *
 	free(x_cpy);
 	free(y_cpy);
 
-	RCTV0 = Ball_Information_wrapper(n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, weight, thread);
-	*bcov = RCTV0;
-	for (i = 0; i<*R; i++)
+	Ball_Information_wrapper(bcov, n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, thread);
+	if (*R > 0)
 	{
-		// stop permutation if user stop it manually:
-		if (pending_interrupt()) {
-			print_stop_message();
-			break;
+		double bcov_tmp[3], *permuted_bcov_weight0, *permuted_bcov_weight_prob, *permuted_bcov_weight_hhg;
+		permuted_bcov_weight0 = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_prob = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_hhg = (double *)malloc(*R * sizeof(double));
+
+		for (i = 0; i<*R; i++)
+		{
+			// stop permutation if user stop it manually:
+			if (pending_interrupt()) {
+				print_stop_message();
+				break;
+			}
+			resample(i_perm, i_perm_inv, n);
+			Ball_Information_wrapper(bcov_tmp, n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, thread);
+			permuted_bcov_weight0[i] = bcov_tmp[0]; permuted_bcov_weight_prob[i] = bcov_tmp[1]; permuted_bcov_weight_hhg[i] = bcov_tmp[2];
 		}
-		resample(i_perm, i_perm_inv, n);
-		permuted_bcov[i] = Ball_Information_wrapper(n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, weight, thread);
+		pvalue[0] = compute_pvalue(bcov[0], permuted_bcov_weight0, i);
+		pvalue[1] = compute_pvalue(bcov[1], permuted_bcov_weight_prob, i);
+		pvalue[2] = compute_pvalue(bcov[2], permuted_bcov_weight_hhg, i);
+
+		free(permuted_bcov_weight0); free(permuted_bcov_weight_prob); free(permuted_bcov_weight_hhg);
 	}
 
 	free_matrix(Dx, *n, *n);
@@ -413,10 +390,10 @@ void BI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *
 }
 
 
-void BI_parallel(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *R, int *weight, int *thread)
+void BI_parallel(double *bcov, double *pvalue, double *x, double *y, int *n, int *R, int *thread)
 {
 	int    i, j, **xidx, **yidx, *i_perm, *i_perm_inv;
-	double **Dx, **Dy, *x_cpy, *y_cpy, RCTV0;
+	double **Dx, **Dy, *x_cpy, *y_cpy;
 
 	Dx = alloc_matrix(*n, *n);
 	Dy = alloc_matrix(*n, *n);
@@ -452,53 +429,66 @@ void BI_parallel(double *bcov, double *permuted_bcov, double *x, double *y, int
 	free(x_cpy);
 	free(y_cpy);
 
-	RCTV0 = Ball_Information(n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, weight);
-	*bcov = RCTV0;
+	Ball_Information(bcov, n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv);
 
 	free_int_matrix(xidx, *n, *n);
 	free_int_matrix(yidx, *n, *n);
 	free(i_perm);
 	free(i_perm_inv);
 
+	if (*R > 0) {
+		double *permuted_bcov_weight0, *permuted_bcov_weight_prob, *permuted_bcov_weight_hhg;
+		permuted_bcov_weight0 = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_prob = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_hhg = (double *)malloc(*R * sizeof(double));
+
 #ifdef _OPENMP
-	omp_set_num_threads(*thread);
+		omp_set_num_threads(*thread);
 #endif
 #pragma omp parallel
-	{
-		int **xidx_thread, **yidx_thread, *i_perm_thread, *i_perm_inv_thread;
-		int k, kk, i_thread;
-		double ans = 0.0;
-		xidx_thread = alloc_int_matrix(*n, *n);
-		yidx_thread = alloc_int_matrix(*n, *n);
-		i_perm_thread = (int *)malloc(*n * sizeof(int));
-		i_perm_inv_thread = (int *)malloc(*n * sizeof(int));
-#pragma omp critical
 		{
-			for (k = 0; k < *n; k++)
+			int **xidx_thread, **yidx_thread, *i_perm_thread, *i_perm_inv_thread;
+			int k, kk, i_thread;
+			double bcov_tmp[3];
+			xidx_thread = alloc_int_matrix(*n, *n);
+			yidx_thread = alloc_int_matrix(*n, *n);
+			i_perm_thread = (int *)malloc(*n * sizeof(int));
+			i_perm_inv_thread = (int *)malloc(*n * sizeof(int));
+#pragma omp critical
 			{
-				for (kk = 0; kk < *n; kk++)
+				for (k = 0; k < *n; k++)
 				{
-					xidx_thread[k][kk] = kk;
-					yidx_thread[k][kk] = kk;
+					for (kk = 0; kk < *n; kk++)
+					{
+						xidx_thread[k][kk] = kk;
+						yidx_thread[k][kk] = kk;
+					}
+					i_perm_thread[k] = k;
+					i_perm_inv_thread[k] = k;
 				}
-				i_perm_thread[k] = k;
-				i_perm_inv_thread[k] = k;
 			}
-		}
+
 #pragma omp for
-		for (i_thread = 0; i_thread < (*R); i_thread++)
-		{
-#pragma omp critical
+			for (i_thread = 0; i_thread < (*R); i_thread++)
 			{
-				resample(i_perm_thread, i_perm_inv_thread, n);
+#pragma omp critical
+				{
+					resample(i_perm_thread, i_perm_inv_thread, n);
+				}
+				Ball_Information(bcov_tmp, n, Dx, Dy, xidx_thread, yidx_thread, i_perm_thread, i_perm_inv_thread);
+				permuted_bcov_weight0[i_thread] = bcov_tmp[0]; 
+				permuted_bcov_weight_prob[i_thread] = bcov_tmp[1]; 
+				permuted_bcov_weight_hhg[i_thread] = bcov_tmp[2];
 			}
-			ans = Ball_Information(n, Dx, Dy, xidx_thread, yidx_thread, i_perm_thread, i_perm_inv_thread, weight);
-			permuted_bcov[i_thread] = ans;
+			free(i_perm_thread);
+			free(i_perm_inv_thread);
+			free_int_matrix(xidx_thread, *n, *n);
+			free_int_matrix(yidx_thread, *n, *n);
 		}
-		free(i_perm_thread);
-		free(i_perm_inv_thread);
-		free_int_matrix(xidx_thread, *n, *n);
-		free_int_matrix(yidx_thread, *n, *n);
+		pvalue[0] = compute_pvalue(bcov[0], permuted_bcov_weight0, *R);
+		pvalue[1] = compute_pvalue(bcov[1], permuted_bcov_weight_prob, *R);
+		pvalue[2] = compute_pvalue(bcov[2], permuted_bcov_weight_hhg, *R);
+		free(permuted_bcov_weight0); free(permuted_bcov_weight_prob); free(permuted_bcov_weight_hhg);
 	}
 
 	free_matrix(Dx, *n, *n);
@@ -507,186 +497,12 @@ void BI_parallel(double *bcov, double *permuted_bcov, double *x, double *y, int
 }
 
 
-void sort(int *n, int *zidx, double *z, int **dzidx)
-{
-	// the z[i] is the center
-	int i, j, zi, ileft, iright, lastpos;
-	double lastval, tmp;
-	for (i = 0; i < *n; i++) {
-		zi = zidx[i];
-		j = *n - 1;
-		lastpos = *n - 1;
-		ileft = 0;
-		iright = *n - 1;
-		lastval = -1.0;
-
-		while ((ileft != i) || (iright != i)) {
-			if (i == ileft) {
-				tmp = z[iright] - z[i];
-				if (lastval != tmp)
-					lastpos = j;
-				dzidx[zi][zidx[iright]] = lastpos;
-				lastval = tmp;
-				iright--;
-			}
-			else if (i == iright) {
-				tmp = z[i] - z[ileft];
-				if (lastval != tmp)
-					lastpos = j;
-				dzidx[zi][zidx[ileft]] = lastpos;
-				lastval = tmp;
-				ileft++;
-			}
-			else {
-				if (z[i] - z[ileft]>z[iright] - z[i]) {
-					tmp = z[i] - z[ileft];
-					if (lastval != tmp)
-						lastpos = j;
-					dzidx[zi][zidx[ileft]] = lastpos;
-					lastval = tmp;
-					ileft++;
-				}
-				else {
-					tmp = z[iright] - z[i];
-					if (lastval != tmp)
-						lastpos = j;
-					dzidx[zi][zidx[iright]] = lastpos;
-					lastval = tmp;
-					iright--;
-				}
-			}
-			j--;
-		}
-
-		if (lastval == 0)
-			dzidx[zi][zi] = lastpos;
-		else
-			dzidx[zi][zi] = 0;
-	}
-}
-
-
-/*
-The algorithm like the quick-sort. It finds the upper and lower in turn. 
-For each i, the computation complexity is O(n); 
-Thus, this function take O(n^2) times.
-
-Algorithm detail:
-For i=1, ..., n, carry out:
-init the low and upper index: l=1, u=N
-while (l <= u):
-if (z[u] - z[i]) >= (z[i] - z[l]) ==> for pair (i, u), upper index: u, lower index: l ==> update upper index: u = u - 1
-if (z[u] - z[i]) < (z[i] - z[l]) ==> for pair (i, l), upper index: u, lower index: l ==> update lower index: l = l - 1
-
-Input: n=6, z = [1, 2, 3, 4, 5, 5], zidx = [3, 1, 5, 2, 6, 4], lowzidx = [], higzidx = [];
-Output: lowzidx, higzidx;
-*/
-void createidx(int *n, int *zidx, double *z, int **lowzidx, int **higzidx)
-{
-	int i, zi, ileft, iright, jleft, jright, lowpos, higpos;
-	double lastval, tmp1, tmp2, tmp;
-	for (i = 0; i < *n; i++) {
-		zi = zidx[i];
-		lowpos = 1;
-		higpos = *n;
-		ileft = 0;
-		iright = *n - 1;
-		jleft = 0;
-		jright = 0;
-
-		tmp1 = z[iright] - z[i];
-		tmp2 = z[i] - z[ileft];
-		if (tmp1 > tmp2) {
-			lastval = tmp1;
-			lowzidx[zi][zidx[iright]] = lowpos;
-			higzidx[zi][zidx[iright]] = higpos;
-			iright--;
-			jright++;
-		}
-		else {
-			lastval = tmp2;
-			if (ileft == i) {
-				lowzidx[zi][zidx[iright]] = lowpos;
-				higzidx[zi][zidx[iright]] = higpos;
-				iright--;
-				jright++;
-			}
-			else {
-				lowzidx[zi][zidx[ileft]] = lowpos;
-				higzidx[zi][zidx[ileft]] = higpos;
-				ileft++;
-				jleft++;
-			}
-		}
-
-		while (ileft <= iright) {
-			tmp1 = z[iright] - z[i];
-			tmp2 = z[i] - z[ileft];
-			tmp = MAX(tmp1, tmp2);
-			while (lastval == tmp) {
-				if (tmp1 > tmp2) {
-					lowzidx[zi][zidx[iright]] = lowpos;
-					higzidx[zi][zidx[iright]] = higpos;
-					iright--;
-					jright++;
-				}
-				else {
-					if (ileft == i) {
-						lowzidx[zi][zidx[iright]] = lowpos;
-						higzidx[zi][zidx[iright]] = higpos;
-						iright--;
-						jright++;
-					}
-					else {
-						lowzidx[zi][zidx[ileft]] = lowpos;
-						higzidx[zi][zidx[ileft]] = higpos;
-						ileft++;
-						jleft++;
-					}
-				}
-				if (iright < ileft)
-					break;
-				tmp1 = z[iright] - z[i];
-				tmp2 = z[i] - z[ileft];
-				tmp = MAX(tmp1, tmp2);
-			}
-			if (iright < ileft)
-				break;
-			lowpos += jleft;
-			higpos -= jright;
-			jleft = 0;
-			jright = 0;
-			if (tmp1 > tmp2) {
-				lastval = tmp1;
-				lowzidx[zi][zidx[iright]] = lowpos;
-				higzidx[zi][zidx[iright]] = higpos;
-				iright--;
-				jright++;
-			}
-			else {
-				lastval = tmp2;
-				if (ileft == i) {
-					lowzidx[zi][zidx[iright]] = lowpos;
-					higzidx[zi][zidx[iright]] = higpos;
-					iright--;
-					jright++;
-				}
-				else {
-					lowzidx[zi][zidx[ileft]] = lowpos;
-					higzidx[zi][zidx[ileft]] = higpos;
-					ileft++;
-					jleft++;
-				}
-			}
-		}
-	}
-}
-
-
-double U_Ball_Information(int *n, int **Rank, int **lowxidx, int **higxidx, int **lowyidx, int **higyidx, int *i_perm, int *weight)
+void U_Ball_Information(double *bcov_stat, int *n, int **Rank, int **lowxidx, int **higxidx, int **lowyidx, int **higyidx, int *i_perm)
 {
 	int i, j, pi, pj;
-	double px, py, pxy, ans = 0;
+	double px, py, pxy;
+	double bcov_weight0 = 0.0, bcov_weight_prob = 0.0, bcov_weight_hhg = 0.0, bcov_fixed_ball = 0.0;
+	double hhg_ball_num = 0.0, minor_ball_prop = 2 / (*n);
 	for (i = 0; i < *n; i++) {
 		for (j = 0; j < *n; j++) {
 			pi = i_perm[i];
@@ -699,27 +515,36 @@ double U_Ball_Information(int *n, int **Rank, int **lowxidx, int **higxidx, int
 			pxy /= (*n);
 			px /= (*n);
 			py /= (*n);
-			if (*weight == 0)
-				ans += pow(pxy - px*py, 2);
-			else
-				ans += pow(pxy - px*py, 2) / (px*py);
+			bcov_fixed_ball = pow(pxy - px*py, 2);
+			bcov_weight0 += bcov_fixed_ball;
+			bcov_weight_prob += bcov_fixed_ball / (px*py);
+			if (px > minor_ball_prop && py > minor_ball_prop && px != 1 && py != 1)
+			{
+				bcov_weight_hhg += bcov_fixed_ball / ((px - minor_ball_prop)*(1.0 - px + minor_ball_prop)*(py - minor_ball_prop)*(1.0 - py + minor_ball_prop));
+				hhg_ball_num += 1;
+			}
 		}
 	}
-	ans /= (1.0*(*n)*(*n));
-	return(ans);
+	bcov_stat[0] = bcov_weight0 / (1.0*(*n)*(*n));
+	bcov_stat[1] = bcov_weight_prob / (1.0*(*n)*(*n));
+	bcov_stat[2] = bcov_weight_hhg / hhg_ball_num;
+
+	return;
 }
 
 
-double U_Ball_Information_parallel(int *n, int **Rank, int **lowxidx, int **higxidx, int **lowyidx, int **higyidx, int *i_perm, int *weight, int *nthread)
+void U_Ball_Information_parallel(double *bcov_stat, int *n, int **Rank, int **lowxidx, int **higxidx, int **lowyidx, int **higyidx, int *i_perm, int *nthread)
 {
-	double BI_value = 0.0;
+	double bcov_weight0 = 0.0, bcov_weight_prob = 0.0, bcov_weight_hhg = 0.0;
+	double minor_ball_prop = 2 / (*n);
 #ifdef _OPENMP
 	omp_set_num_threads(*nthread);
 #endif
 #pragma omp parallel
 	{
 		int i, j, pi, pj;
-		double px, py, pxy, ans = 0;
+		double px, py, pxy;
+		double bcov_weight0_thread = 0.0, bcov_weight_prob_thread = 0.0, bcov_weight_hhg_thread = 0.0, bcov_fixed_ball = 0.0;
 
 #pragma omp for
 		for (i = 0; i < *n; i++) {
@@ -734,39 +559,45 @@ double U_Ball_Information_parallel(int *n, int **Rank, int **lowxidx, int **higx
 				pxy /= (*n);
 				px /= (*n);
 				py /= (*n);
-				if (*weight == 0)
-					ans += pow(pxy - px*py, 2);
-				else
-					ans += pow(pxy - px*py, 2) / (px*py);
+
+				bcov_fixed_ball = (pxy - px*py) * (pxy - px*py);
+				bcov_weight0_thread += bcov_fixed_ball;
+				bcov_weight_prob_thread += bcov_fixed_ball / (px*py);
+				if (px > minor_ball_prop && py > minor_ball_prop && px != 1 && py != 1)
+				{
+					bcov_weight_hhg_thread += bcov_fixed_ball / ((px - minor_ball_prop)*(1.0 - px + minor_ball_prop)*(py - minor_ball_prop)*(1.0 - py + minor_ball_prop));
+				}
 			}
 		}
 
 #pragma omp critical
 		{
-			BI_value += ans;
+			bcov_weight0 += bcov_weight0_thread;
+			bcov_weight_prob += bcov_weight_prob_thread;
+			bcov_weight_hhg += bcov_weight_hhg_thread;
 		}
 	}
-
-	BI_value /= (1.0*(*n)*(*n));
-	return(BI_value);
+	bcov_stat[0] = bcov_weight0;
+	bcov_stat[1] = bcov_weight_prob;
+	bcov_stat[2] = bcov_weight_hhg;
+	return;
 }
 
 
-double U_Ball_Information_wrapper(int *n, int **Rank, int **lowxidx, int **higxidx, int **lowyidx, int **higyidx, int *i_perm, int *weight, int *nthread)
+void U_Ball_Information_wrapper(double *bcov_stat, int *n, int **Rank, int **lowxidx, int **higxidx, int **lowyidx, int **higyidx, int *i_perm, int *nthread)
 {
-	double ans = 0.0;
 	if ((*nthread) == 1)
 	{
-		ans = U_Ball_Information(n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, weight);
+		U_Ball_Information(bcov_stat, n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm);
 	}
 	else {
-		ans = U_Ball_Information_parallel(n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, weight, nthread);
+		U_Ball_Information_parallel(bcov_stat, n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, nthread);
 	}
-	return ans;
+	return;
 }
 
 
-void UBI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *R, int *weight, int *thread)
+void UBI(double *bcov, double *pvalue, double *x, double *y, int *n, int *R, int *thread)
 {
 	int i, j, *xidx, *yidx, *xrank, *yrank, *i_perm, **Rank, **lowxidx, **higxidx, **lowyidx, **higyidx;
 
@@ -798,17 +629,31 @@ void UBI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int
 
 
 	initRank(*n, Rank, xrank, yrank, i_perm);
-	*bcov = U_Ball_Information_wrapper(n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, weight, thread);
-	for (j = 0; j<*R; j++)
+	U_Ball_Information_wrapper(bcov, n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, thread);
+	if (*R > 0)
 	{
-		// stop permutation if user stop it manually:
-		if (pending_interrupt()) {
-			print_stop_message();
-			break;
+		double bcov_tmp[3], *permuted_bcov_weight0, *permuted_bcov_weight_prob, *permuted_bcov_weight_hhg;
+		permuted_bcov_weight0 = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_prob = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_hhg = (double *)malloc(*R * sizeof(double));
+
+		for (j = 0; j<*R; j++)
+		{
+			// stop permutation if user stop it manually:
+			if (pending_interrupt()) {
+				print_stop_message();
+				break;
+			}
+			resample2(i_perm, n);
+			initRank(*n, Rank, xrank, yrank, i_perm);
+			U_Ball_Information_wrapper(bcov_tmp, n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, thread);
+			permuted_bcov_weight0[j] = bcov_tmp[0]; permuted_bcov_weight_prob[j] = bcov_tmp[1]; permuted_bcov_weight_hhg[j] = bcov_tmp[2];
 		}
-		resample2(i_perm, n);
-		initRank(*n, Rank, xrank, yrank, i_perm);
-		permuted_bcov[j] = U_Ball_Information_wrapper(n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, weight, thread);
+		pvalue[0] = compute_pvalue(bcov[0], permuted_bcov_weight0, j);
+		pvalue[1] = compute_pvalue(bcov[1], permuted_bcov_weight_prob, j);
+		pvalue[2] = compute_pvalue(bcov[2], permuted_bcov_weight_hhg, j);
+
+		free(permuted_bcov_weight0); free(permuted_bcov_weight_prob); free(permuted_bcov_weight_hhg);
 	}
 
 	free_int_matrix(Rank, (*n) + 1, (*n) + 1);
@@ -825,7 +670,7 @@ void UBI(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int
 }
 
 
-void UBI_parallel(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *R, int *weight, int *thread)
+void UBI_parallel(double *bcov, double *pvalue, double *x, double *y, int *n, int *R, int *thread)
 {
 	int i, *xidx, *yidx, *xrank, *yrank, *i_perm, **Rank, **lowxidx, **higxidx, **lowyidx, **higyidx;
 
@@ -855,47 +700,63 @@ void UBI_parallel(double *bcov, double *permuted_bcov, double *x, double *y, int
 	createidx(n, xidx, x, lowxidx, higxidx);
 	createidx(n, yidx, y, lowyidx, higyidx);
 
-
 	initRank(*n, Rank, xrank, yrank, i_perm);
-	*bcov = U_Ball_Information(n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, weight);
+	U_Ball_Information(bcov, n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm);
 
 	free_int_matrix(Rank, (*n) + 1, (*n) + 1);
 	free(i_perm);
 	free(xidx);
 	free(yidx);
 
+	if (*R > 0)
+	{
+		double *permuted_bcov_weight0, *permuted_bcov_weight_prob, *permuted_bcov_weight_hhg;
+		permuted_bcov_weight0 = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_prob = (double *)malloc(*R * sizeof(double));
+		permuted_bcov_weight_hhg = (double *)malloc(*R * sizeof(double));
+
 #ifdef _OPENMP
-	omp_set_num_threads(*thread);
+		omp_set_num_threads(*thread);
 #endif
 #pragma omp parallel
-	{
-		int **Rank_thread, *i_perm_thread;
-		int k, j_thread;
-		double ans_thread = 0.0;
-		i_perm_thread = (int *)malloc(*n * sizeof(int));
-		Rank_thread = alloc_int_matrix((*n) + 1, (*n) + 1);
-#pragma omp critical
 		{
-			for (k = 0; k < *n; k++)
-			{
-				i_perm_thread[k] = k;
-			}
-		}
-#pragma omp for
-		for (j_thread = 0; j_thread < (*R); j_thread++) {
+			int **Rank_thread, *i_perm_thread;
+			int k, j_thread;
+			double bcov_tmp[3];
+			i_perm_thread = (int *)malloc(*n * sizeof(int));
+			Rank_thread = alloc_int_matrix((*n) + 1, (*n) + 1);
 #pragma omp critical
 			{
-				// TODO: I don't know why this command can only be runned in a single thread. But it makes R console temporarily available and seems to be correct.
-				resample2(i_perm_thread, n);
+				for (k = 0; k < *n; k++)
+				{
+					i_perm_thread[k] = k;
+				}
 			}
-			initRank(*n, Rank_thread, xrank, yrank, i_perm_thread);
-			ans_thread = U_Ball_Information(n, Rank_thread, lowxidx, higxidx, lowyidx, higyidx, i_perm_thread, weight);
-			permuted_bcov[j_thread] = ans_thread;
+#pragma omp for
+			for (j_thread = 0; j_thread < (*R); j_thread++) {
+#pragma omp critical
+				{
+					// TODO: I don't know why this command can only be runned in a single thread. But it makes R console temporarily available and seems to be correct.
+					resample2(i_perm_thread, n);
+				}
+				initRank(*n, Rank_thread, xrank, yrank, i_perm_thread);
+				U_Ball_Information(bcov_tmp, n, Rank_thread, lowxidx, higxidx, lowyidx, higyidx, i_perm_thread);
+				permuted_bcov_weight0[j_thread] = bcov_tmp[0];
+				permuted_bcov_weight_prob[j_thread] = bcov_tmp[1];
+				permuted_bcov_weight_hhg[j_thread] = bcov_tmp[2];
+			}
+			free(i_perm_thread);
+			free_int_matrix(Rank_thread, (*n) + 1, (*n) + 1);
 		}
-		free(i_perm_thread);
-		free_int_matrix(Rank_thread, (*n) + 1, (*n) + 1);
+
+		pvalue[0] = compute_pvalue(bcov[0], permuted_bcov_weight0, *R);
+		pvalue[1] = compute_pvalue(bcov[1], permuted_bcov_weight_prob, *R);
+		pvalue[2] = compute_pvalue(bcov[2], permuted_bcov_weight_hhg, *R);
+
+		free(permuted_bcov_weight0); free(permuted_bcov_weight_prob); free(permuted_bcov_weight_hhg);
 	}
 
+
 	free_int_matrix(lowxidx, *n, *n);
 	free_int_matrix(higxidx, *n, *n);
 	free_int_matrix(lowyidx, *n, *n);
@@ -907,180 +768,13 @@ void UBI_parallel(double *bcov, double *permuted_bcov, double *x, double *y, int
 }
 
 
-double ubcov_value(double *x, double *y, int *n, int *weight, int *thread)
-{
-	int    i, *xidx, *yidx, *xrank, *yrank, *i_perm, **Rank, **lowxidx, **higxidx, **lowyidx, **higyidx;
-	double RCTV0;
-
-	xidx = (int *)malloc(*n * sizeof(int));
-	yidx = (int *)malloc(*n * sizeof(int));
-	xrank = (int *)malloc(*n * sizeof(int));
-	yrank = (int *)malloc(*n * sizeof(int));
-	i_perm = (int *)malloc(*n * sizeof(int));
-	Rank = alloc_int_matrix((*n) + 1, (*n) + 1);
-	lowxidx = alloc_int_matrix(*n, *n);
-	higxidx = alloc_int_matrix(*n, *n);
-	lowyidx = alloc_int_matrix(*n, *n);
-	higyidx = alloc_int_matrix(*n, *n);
-
-	for (i = 0; i<*n; i++)
-	{
-		xidx[i] = i;
-		yidx[i] = i;
-		i_perm[i] = i;
-	}
-
-	// first step: sort the x
-	quicksort(x, xidx, 0, *n - 1);
-	quicksort(y, yidx, 0, *n - 1);
-	ranksort(n, xrank, x, xidx);
-	ranksort(n, yrank, y, yidx);
-	createidx(n, xidx, x, lowxidx, higxidx);
-	createidx(n, yidx, y, lowyidx, higyidx);
-
-
-	initRank(*n, Rank, xrank, yrank, i_perm);
-	RCTV0 = U_Ball_Information_wrapper(n, Rank, lowxidx, higxidx, lowyidx, higyidx, i_perm, weight, thread);
-
-	free_int_matrix(Rank, (*n) + 1, (*n) + 1);
-	free_int_matrix(lowxidx, *n, *n);
-	free_int_matrix(higxidx, *n, *n);
-	free_int_matrix(lowyidx, *n, *n);
-	free_int_matrix(higyidx, *n, *n);
-	free(xidx);
-	free(yidx);
-	free(xrank);
-	free(yrank);
-	free(i_perm);
-	return(RCTV0);
-}
-
-
-double bcov_value(double *x, double *y, int *n, int *weight, int *thread)
-{
-	/*  computes RCT(x,y)  */
-	int    i, j, **xidx, **yidx, *i_perm, *i_perm_inv;
-	double **Dx, **Dy, *x_cpy, *y_cpy, RCTV0;
-
-	Dx = alloc_matrix(*n, *n);
-	Dy = alloc_matrix(*n, *n);
-	xidx = alloc_int_matrix(*n, *n);
-	yidx = alloc_int_matrix(*n, *n);
-	i_perm = (int *)malloc(*n * sizeof(int));
-	i_perm_inv = (int *)malloc(*n * sizeof(int));
-	x_cpy = (double *)malloc(*n * sizeof(double));
-	y_cpy = (double *)malloc(*n * sizeof(double));
-
-	vector2matrix(x, Dx, *n, *n, 1);
-	vector2matrix(y, Dy, *n, *n, 1);
-
-	for (i = 0; i<*n; i++)
-	{
-		for (j = 0; j<*n; j++)
-		{
-			xidx[i][j] = j;
-			yidx[i][j] = j;
-		}
-		i_perm[i] = i;
-		i_perm_inv[i] = i;
-	}
-
-	for (i = 0; i<(*n); i++)
-	{
-		// copy site to x_cpy and y_cpy
-		memcpy(x_cpy, Dx[i], *n * sizeof(double));
-		memcpy(y_cpy, Dy[i], *n * sizeof(double));
-		quicksort(x_cpy, xidx[i], 0, *n - 1);
-		quicksort(y_cpy, yidx[i], 0, *n - 1);
-	}
-	free(x_cpy);
-	free(y_cpy);
-
-	RCTV0 = Ball_Information_wrapper(n, Dx, Dy, xidx, yidx, i_perm, i_perm_inv, weight, thread);
-
-	free_matrix(Dx, *n, *n);
-	free_matrix(Dy, *n, *n);
-	free_int_matrix(xidx, *n, *n);
-	free_int_matrix(yidx, *n, *n);
-	free(i_perm);
-	free(i_perm_inv);
-	return(RCTV0);
-}
-
-
-double bcor_value(double *x, double *y, int *n, int *weight, int *dst, int *thread)
-{
-	double bcov_stat_xy, bcov_stat_xx, bcov_stat_yy;
-	double bcor_stat;
-	if ((*dst)) {
-		bcov_stat_xy = bcov_value(x, y, n, weight, thread);
-		bcov_stat_xx = bcov_value(x, x, n, weight, thread);
-		bcov_stat_yy = bcov_value(x, x, n, weight, thread);
-	}
-	else {
-		bcov_stat_xy = ubcov_value(x, y, n, weight, thread);
-		bcov_stat_xx = ubcov_value(x, x, n, weight, thread);
-		bcov_stat_yy = ubcov_value(x, y, n, weight, thread);
-	}
-	bcor_stat = bcov_stat_xy / sqrt(bcov_stat_xx*bcov_stat_yy);
-	return(bcor_stat);
-}
-
-
-void UBCOR(double *bcor, double *permuted_bcor, double *x, double *y, int *n, int *R, int *weight, int *thread)
-{
-	int i;
-	double bcov_stat_xx, bcov_stat_yy;
-	UBI(bcor, permuted_bcor, x, y, n, R, weight, thread);
-	bcov_stat_xx = ubcov_value(x, x, n, weight, thread);
-	bcov_stat_yy = ubcov_value(y, y, n, weight, thread);
-	(*bcor) = (*bcor) / bcov_stat_xx / bcov_stat_yy;
-	for (i = 0; i < (*R); i++) {
-		permuted_bcor[i] = permuted_bcor[i] / bcov_stat_xx / bcov_stat_yy;
-	}
-}
-
-
-void BCOR(double *bcor, double *permuted_bcor, double *x, double *y, int *n, int *R, int *weight, int *thread)
-{
-	int i;
-	double bcov_stat_xx, bcov_stat_yy;
-	BI(bcor, permuted_bcor, x, y, n, R, weight, thread);
-	bcov_stat_xx = bcov_value(x, x, n, weight, thread);
-	bcov_stat_yy = bcov_value(y, y, n, weight, thread);
-	(*bcor) = (*bcor) / bcov_stat_xx / bcov_stat_yy;
-	for (i = 0; i < (*R); i++) {
-		permuted_bcor[i] = permuted_bcor[i] / bcov_stat_xx / bcov_stat_yy;
-	}
-}
-
-
 /////////////////////////////////////////////////////////////////
 //////////// two functions below wrap key function //////////////
 // bcov_stat return ball covariance statistic 
 // bcov_test execute ball covariance statistic based test
 /////////////////////////////////////////////////////////////////
 
-void bcov_stat(double *bcov, double *x, double *y, int *n, int *weight, int *dst, int *type, int *thread)
-{
-	double bcov_stat;
-	if ((*type) == 1) {
-		if ((*dst)) {
-			bcov_stat = bcov_value(x, y, n, weight, thread);
-		}
-		else {
-			bcov_stat = ubcov_value(x, y, n, weight, thread);
-		}
-		*bcov = bcov_stat;
-	}
-	else {
-		*bcov = bcor_value(x, y, n, weight, dst, thread);
-	}
-	return;
-}
-
-
-void bcov_test(double *bcov, double *permuted_bcov, double *x, double *y, int *n, int *R, int *weight, int *dst, int *type, int *thread)
+void bcov_test(double *bcov, double *pvalue, double *x, double *y, int *n, int *R, int *dst, int *thread)
 {
 	//parallel method
 	// if parallel_type == 1, we parallel the computation through statistics.
@@ -1094,33 +788,23 @@ void bcov_test(double *bcov, double *permuted_bcov, double *x, double *y, int *n
 	{
 		*thread = 1;
 	}
-	if ((*type) == 1) {
-		if ((*dst)) {
-			if (parallel_type == 2)
-			{
-				BI_parallel(bcov, permuted_bcov, x, y, n, R, weight, thread);
-			}
-			else {
-				BI(bcov, permuted_bcov, x, y, n, R, weight, thread);
-			}
+	if ((*dst)) {
+		if (parallel_type == 2)
+		{
+			BI_parallel(bcov, pvalue, x, y, n, R, thread);
 		}
 		else {
-			if (parallel_type == 2)
-			{
-				UBI_parallel(bcov, permuted_bcov, x, y, n, R, weight, thread);
-			}
-			else
-			{
-				UBI(bcov, permuted_bcov, x, y, n, R, weight, thread);
-			}
+			BI(bcov, pvalue, x, y, n, R, thread);
 		}
 	}
 	else {
-		if ((*dst)) {
-			BCOR(bcov, permuted_bcov, x, y, n, R, weight, thread);
+		if (parallel_type == 2)
+		{
+			UBI_parallel(bcov, pvalue, x, y, n, R, thread);
 		}
-		else {
-			UBCOR(bcov, permuted_bcov, x, y, n, R, weight, thread);
+		else
+		{
+			UBI(bcov, pvalue, x, y, n, R, thread);
 		}
 	}
 	return;
